@@ -3,12 +3,12 @@
  * nodeIndexscan.c
  *	  Routines to support indexes and indexed scans of relations
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/executor/nodeIndexscan.c,v 1.71 2002/09/04 20:31:18 momjian Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/executor/nodeIndexscan.c,v 1.84 2003/09/24 18:54:01 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -28,19 +28,51 @@
 #include "access/heapam.h"
 #include "executor/execdebug.h"
 #include "executor/nodeIndexscan.h"
+#include "miscadmin.h"
 #include "nodes/nodeFuncs.h"
 #include "optimizer/clauses.h"
 #include "parser/parsetree.h"
 
-/* ----------------
- *		Misc stuff to move to executor.h soon -cim 6/5/90
- * ----------------
- */
+
 #define NO_OP			0
 #define LEFT_OP			1
 #define RIGHT_OP		2
 
-static TupleTableSlot *IndexNext(IndexScan *node);
+/*
+ * In a multiple-index plan, we must take care to return any given tuple
+ * only once, even if it matches conditions of several index scans.  Our
+ * preferred way to do this is to record already-returned tuples in a hash
+ * table (using the TID as unique identifier).  However, in a very large
+ * scan this could conceivably run out of memory.  We limit the hash table
+ * to no more than SortMem KB; if it grows past that, we fall back to the
+ * pre-7.4 technique: evaluate the prior-scan index quals again for each
+ * tuple (which is space-efficient, but slow).
+ *
+ * When scanning backwards, we use scannum to determine when to emit the
+ * tuple --- we have to re-emit a tuple in the same scan as it was first
+ * encountered.
+ *
+ * Note: this code would break if the planner were ever to create a multiple
+ * index plan with overall backwards direction, because the hashtable code
+ * will emit a tuple the first time it is encountered (which would be the
+ * highest scan in which it matches the index), but the evaluate-the-quals
+ * code will emit a tuple in the lowest-numbered scan in which it's valid.
+ * This could be fixed at need by making the evaluate-the-quals case more
+ * complex.  Currently the planner will never create such a plan (since it
+ * considers multi-index plans unordered anyway), so there's no need for
+ * more complexity.
+ */
+typedef struct
+{
+	/* tid is the hash key and so must be first! */
+	ItemPointerData tid;		/* TID of a tuple we've returned */
+	int			scannum;		/* number of scan we returned it in */
+} DupHashTabEntry;
+
+
+static TupleTableSlot *IndexNext(IndexScanState *node);
+static void create_duphash(IndexScanState *node);
+
 
 /* ----------------------------------------------------------------
  *		IndexNext
@@ -65,15 +97,14 @@ static TupleTableSlot *IndexNext(IndexScan *node);
  * ----------------------------------------------------------------
  */
 static TupleTableSlot *
-IndexNext(IndexScan *node)
+IndexNext(IndexScanState *node)
 {
 	EState	   *estate;
-	CommonScanState *scanstate;
-	IndexScanState *indexstate;
 	ExprContext *econtext;
 	ScanDirection direction;
 	IndexScanDescPtr scanDescs;
 	IndexScanDesc scandesc;
+	Index		scanrelid;
 	HeapTuple	tuple;
 	TupleTableSlot *slot;
 	int			numIndices;
@@ -83,21 +114,20 @@ IndexNext(IndexScan *node)
 	/*
 	 * extract necessary information from index scan node
 	 */
-	estate = node->scan.plan.state;
+	estate = node->ss.ps.state;
 	direction = estate->es_direction;
-	if (ScanDirectionIsBackward(node->indxorderdir))
+	if (ScanDirectionIsBackward(((IndexScan *) node->ss.ps.plan)->indxorderdir))
 	{
 		if (ScanDirectionIsForward(direction))
 			direction = BackwardScanDirection;
 		else if (ScanDirectionIsBackward(direction))
 			direction = ForwardScanDirection;
 	}
-	scanstate = node->scan.scanstate;
-	indexstate = node->indxstate;
-	scanDescs = indexstate->iss_ScanDescs;
-	numIndices = indexstate->iss_NumIndices;
-	econtext = scanstate->cstate.cs_ExprContext;
-	slot = scanstate->css_ScanTupleSlot;
+	scanDescs = node->iss_ScanDescs;
+	numIndices = node->iss_NumIndices;
+	econtext = node->ss.ps.ps_ExprContext;
+	slot = node->ss.ss_ScanTupleSlot;
+	scanrelid = ((IndexScan *) node->ss.ps.plan)->scan.scanrelid;
 
 	/*
 	 * Check if we are evaluating PlanQual for tuple of this relation.
@@ -106,15 +136,15 @@ IndexNext(IndexScan *node)
 	 * switching in Init/ReScan plan...
 	 */
 	if (estate->es_evTuple != NULL &&
-		estate->es_evTuple[node->scan.scanrelid - 1] != NULL)
+		estate->es_evTuple[scanrelid - 1] != NULL)
 	{
 		List	   *qual;
 
 		ExecClearTuple(slot);
-		if (estate->es_evTupleNull[node->scan.scanrelid - 1])
+		if (estate->es_evTupleNull[scanrelid - 1])
 			return slot;		/* return empty slot */
 
-		ExecStoreTuple(estate->es_evTuple[node->scan.scanrelid - 1],
+		ExecStoreTuple(estate->es_evTuple[scanrelid - 1],
 					   slot, InvalidBuffer, false);
 
 		/* Does the tuple meet any of the OR'd indxqual conditions? */
@@ -131,7 +161,7 @@ IndexNext(IndexScan *node)
 			slot->val = NULL;
 
 		/* Flag for the next call that no more tuples */
-		estate->es_evTupleNull[node->scan.scanrelid - 1] = true;
+		estate->es_evTupleNull[scanrelid - 1] = true;
 
 		return slot;
 	}
@@ -144,28 +174,28 @@ IndexNext(IndexScan *node)
 	bBackward = ScanDirectionIsBackward(direction);
 	if (bBackward)
 	{
-		indexNumber = numIndices - indexstate->iss_IndexPtr - 1;
+		indexNumber = numIndices - node->iss_IndexPtr - 1;
 		if (indexNumber < 0)
 		{
 			indexNumber = 0;
-			indexstate->iss_IndexPtr = numIndices - 1;
+			node->iss_IndexPtr = numIndices - 1;
 		}
 	}
 	else
 	{
-		if ((indexNumber = indexstate->iss_IndexPtr) < 0)
+		if ((indexNumber = node->iss_IndexPtr) < 0)
 		{
 			indexNumber = 0;
-			indexstate->iss_IndexPtr = 0;
+			node->iss_IndexPtr = 0;
 		}
 	}
 	while (indexNumber < numIndices)
 	{
-		scandesc = scanDescs[indexstate->iss_IndexPtr];
+		scandesc = scanDescs[node->iss_IndexPtr];
 		while ((tuple = index_getnext(scandesc, direction)) != NULL)
 		{
 			/*
-			 * store the scanned tuple in the scan tuple slot of the scan
+			 * Store the scanned tuple in the scan tuple slot of the scan
 			 * state.  Note: we pass 'false' because tuples returned by
 			 * amgetnext are pointers onto disk pages and must not be
 			 * pfree()'d.
@@ -176,36 +206,80 @@ IndexNext(IndexScan *node)
 						   false);		/* don't pfree */
 
 			/*
-			 * We must check to see if the current tuple was already
-			 * matched by an earlier index, so we don't double-report it.
-			 * We do this by passing the tuple through ExecQual and
-			 * checking for failure with all previous qualifications.
+			 * If it's a multiple-index scan, make sure not to double-report
+			 * a tuple matched by more than one index.  (See notes above.)
 			 */
-			if (indexstate->iss_IndexPtr > 0)
+			if (numIndices > 1)
 			{
-				bool		prev_matches = false;
-				int			prev_index;
-				List	   *qual;
+				/* First try the hash table */
+				if (node->iss_DupHash)
+				{
+					DupHashTabEntry *entry;
+					bool	found;
 
-				econtext->ecxt_scantuple = slot;
-				ResetExprContext(econtext);
-				qual = node->indxqualorig;
-				for (prev_index = 0;
-					 prev_index < indexstate->iss_IndexPtr;
-					 prev_index++)
-				{
-					if (ExecQual((List *) lfirst(qual), econtext, false))
+					entry = (DupHashTabEntry *)
+						hash_search(node->iss_DupHash,
+									&tuple->t_data->t_ctid,
+									HASH_ENTER,
+									&found);
+					if (entry == NULL ||
+						node->iss_DupHash->hctl->nentries > node->iss_MaxHash)
 					{
-						prev_matches = true;
-						break;
+						/* out of memory (either hard or soft limit) */
+						/* release hash table and fall thru to old code */
+						hash_destroy(node->iss_DupHash);
+						node->iss_DupHash = NULL;
+					}
+					else if (found)
+					{
+						/* pre-existing entry */
+
+						/*
+						 * It's duplicate if first emitted in a different
+						 * scan.  If same scan, we must be backing up, so
+						 * okay to emit again.
+						 */
+						if (entry->scannum != node->iss_IndexPtr)
+						{
+							/* Dup, so drop it and loop back for another */
+							ExecClearTuple(slot);
+							continue;
+						}
+					}
+					else
+					{
+						/* new entry, finish filling it in */
+						entry->scannum = node->iss_IndexPtr;
 					}
-					qual = lnext(qual);
 				}
-				if (prev_matches)
+				/* If hash table has overflowed, do it the hard way */
+				if (node->iss_DupHash == NULL &&
+					node->iss_IndexPtr > 0)
 				{
-					/* Duplicate, so drop it and loop back for another */
-					ExecClearTuple(slot);
-					continue;
+					bool		prev_matches = false;
+					int			prev_index;
+					List	   *qual;
+
+					econtext->ecxt_scantuple = slot;
+					ResetExprContext(econtext);
+					qual = node->indxqualorig;
+					for (prev_index = 0;
+						 prev_index < node->iss_IndexPtr;
+						 prev_index++)
+					{
+						if (ExecQual((List *) lfirst(qual), econtext, false))
+						{
+							prev_matches = true;
+							break;
+						}
+						qual = lnext(qual);
+					}
+					if (prev_matches)
+					{
+						/* Dup, so drop it and loop back for another */
+						ExecClearTuple(slot);
+						continue;
+					}
 				}
 			}
 
@@ -216,9 +290,9 @@ IndexNext(IndexScan *node)
 		{
 			indexNumber++;
 			if (bBackward)
-				indexstate->iss_IndexPtr--;
+				node->iss_IndexPtr--;
 			else
-				indexstate->iss_IndexPtr++;
+				node->iss_IndexPtr++;
 		}
 	}
 
@@ -251,21 +325,19 @@ IndexNext(IndexScan *node)
  * ----------------------------------------------------------------
  */
 TupleTableSlot *
-ExecIndexScan(IndexScan *node)
+ExecIndexScan(IndexScanState *node)
 {
-	IndexScanState *indexstate = node->indxstate;
-
 	/*
 	 * If we have runtime keys and they've not already been set up, do it
 	 * now.
 	 */
-	if (indexstate->iss_RuntimeKeyInfo && !indexstate->iss_RuntimeKeysReady)
-		ExecReScan((Plan *) node, NULL, NULL);
+	if (node->iss_RuntimeKeyInfo && !node->iss_RuntimeKeysReady)
+		ExecReScan((PlanState *) node, NULL);
 
 	/*
 	 * use IndexNext as access method
 	 */
-	return ExecScan(&node->scan, (ExecScanAccessMtd) IndexNext);
+	return ExecScan(&node->ss, (ExecScanAccessMtd) IndexNext);
 }
 
 /* ----------------------------------------------------------------
@@ -280,28 +352,28 @@ ExecIndexScan(IndexScan *node)
  * ----------------------------------------------------------------
  */
 void
-ExecIndexReScan(IndexScan *node, ExprContext *exprCtxt, Plan *parent)
+ExecIndexReScan(IndexScanState *node, ExprContext *exprCtxt)
 {
 	EState	   *estate;
-	IndexScanState *indexstate;
 	ExprContext *econtext;
 	int			numIndices;
 	IndexScanDescPtr scanDescs;
 	ScanKey    *scanKeys;
-	int		  **runtimeKeyInfo;
+	ExprState ***runtimeKeyInfo;
 	int		   *numScanKeys;
+	Index		scanrelid;
 	int			i;
 	int			j;
 
-	estate = node->scan.plan.state;
-	indexstate = node->indxstate;
-	econtext = indexstate->iss_RuntimeContext;	/* context for runtime
+	estate = node->ss.ps.state;
+	econtext = node->iss_RuntimeContext;		/* context for runtime
 												 * keys */
-	numIndices = indexstate->iss_NumIndices;
-	scanDescs = indexstate->iss_ScanDescs;
-	scanKeys = indexstate->iss_ScanKeys;
-	runtimeKeyInfo = indexstate->iss_RuntimeKeyInfo;
-	numScanKeys = indexstate->iss_NumScanKeys;
+	numIndices = node->iss_NumIndices;
+	scanDescs = node->iss_ScanDescs;
+	scanKeys = node->iss_ScanKeys;
+	runtimeKeyInfo = node->iss_RuntimeKeyInfo;
+	numScanKeys = node->iss_NumScanKeys;
+	scanrelid = ((IndexScan *) node->ss.ps.plan)->scan.scanrelid;
 
 	if (econtext)
 	{
@@ -315,7 +387,7 @@ ExecIndexReScan(IndexScan *node, ExprContext *exprCtxt, Plan *parent)
 			ExprContext *stdecontext;
 
 			econtext->ecxt_outertuple = exprCtxt->ecxt_outertuple;
-			stdecontext = node->scan.scanstate->cstate.cs_ExprContext;
+			stdecontext = node->ss.ps.ps_ExprContext;
 			stdecontext->ecxt_outertuple = exprCtxt->ecxt_outertuple;
 		}
 
@@ -333,29 +405,18 @@ ExecIndexReScan(IndexScan *node, ExprContext *exprCtxt, Plan *parent)
 	 */
 	if (runtimeKeyInfo)
 	{
-		List	   *indxqual;
-
-		indxqual = node->indxqual;
 		for (i = 0; i < numIndices; i++)
 		{
-			List	   *qual = lfirst(indxqual);
 			int			n_keys;
 			ScanKey		scan_keys;
-			int		   *run_keys;
-			List	   *listscan;
+			ExprState **run_keys;
 
-			indxqual = lnext(indxqual);
 			n_keys = numScanKeys[i];
 			scan_keys = scanKeys[i];
 			run_keys = runtimeKeyInfo[i];
 
-			listscan = qual;
 			for (j = 0; j < n_keys; j++)
 			{
-				Expr	   *clause = lfirst(listscan);
-
-				listscan = lnext(listscan);
-
 				/*
 				 * If we have a run-time key, then extract the run-time
 				 * expression and evaluate it with respect to the current
@@ -369,17 +430,12 @@ ExecIndexReScan(IndexScan *node, ExprContext *exprCtxt, Plan *parent)
 				 * is wrong, we could copy the result into our context
 				 * explicitly, but I think that's not necessary...
 				 */
-				if (run_keys[j] != NO_OP)
+				if (run_keys[j] != NULL)
 				{
-					Node	   *scanexpr;
 					Datum		scanvalue;
 					bool		isNull;
 
-					scanexpr = (run_keys[j] == RIGHT_OP) ?
-						(Node *) get_rightop(clause) :
-						(Node *) get_leftop(clause);
-
-					scanvalue = ExecEvalExprSwitchContext(scanexpr,
+					scanvalue = ExecEvalExprSwitchContext(run_keys[j],
 														  econtext,
 														  &isNull,
 														  NULL);
@@ -392,22 +448,30 @@ ExecIndexReScan(IndexScan *node, ExprContext *exprCtxt, Plan *parent)
 			}
 		}
 
-		indexstate->iss_RuntimeKeysReady = true;
+		node->iss_RuntimeKeysReady = true;
 	}
 
 	/* If this is re-scanning of PlanQual ... */
 	if (estate->es_evTuple != NULL &&
-		estate->es_evTuple[node->scan.scanrelid - 1] != NULL)
+		estate->es_evTuple[scanrelid - 1] != NULL)
 	{
-		estate->es_evTupleNull[node->scan.scanrelid - 1] = false;
+		estate->es_evTupleNull[scanrelid - 1] = false;
 		return;
 	}
 
+	/* reset hash table */
+	if (numIndices > 1)
+	{
+		if (node->iss_DupHash)
+			hash_destroy(node->iss_DupHash);
+		create_duphash(node);
+	}
+
 	/* reset index scans */
-	if (ScanDirectionIsBackward(node->indxorderdir))
-		indexstate->iss_IndexPtr = numIndices;
+	if (ScanDirectionIsBackward(((IndexScan *) node->ss.ps.plan)->indxorderdir))
+		node->iss_IndexPtr = numIndices;
 	else
-		indexstate->iss_IndexPtr = -1;
+		node->iss_IndexPtr = -1;
 
 	for (i = 0; i < numIndices; i++)
 	{
@@ -420,53 +484,41 @@ ExecIndexReScan(IndexScan *node, ExprContext *exprCtxt, Plan *parent)
 
 /* ----------------------------------------------------------------
  *		ExecEndIndexScan
- *
- * old comments
- *		Releases any storage allocated through C routines.
- *		Returns nothing.
  * ----------------------------------------------------------------
  */
 void
-ExecEndIndexScan(IndexScan *node)
+ExecEndIndexScan(IndexScanState *node)
 {
-	CommonScanState *scanstate;
-	IndexScanState *indexstate;
-	int		  **runtimeKeyInfo;
-	ScanKey    *scanKeys;
-	List	   *indxqual;
-	int		   *numScanKeys;
 	int			numIndices;
-	Relation	relation;
 	RelationPtr indexRelationDescs;
 	IndexScanDescPtr indexScanDescs;
+	Relation	relation;
 	int			i;
 
-	scanstate = node->scan.scanstate;
-	indexstate = node->indxstate;
-	indxqual = node->indxqual;
-	runtimeKeyInfo = indexstate->iss_RuntimeKeyInfo;
-
 	/*
 	 * extract information from the node
 	 */
-	numIndices = indexstate->iss_NumIndices;
-	scanKeys = indexstate->iss_ScanKeys;
-	numScanKeys = indexstate->iss_NumScanKeys;
-	indexRelationDescs = indexstate->iss_RelationDescs;
-	indexScanDescs = indexstate->iss_ScanDescs;
-	relation = scanstate->css_currentRelation;
+	numIndices = node->iss_NumIndices;
+	indexRelationDescs = node->iss_RelationDescs;
+	indexScanDescs = node->iss_ScanDescs;
+	relation = node->ss.ss_currentRelation;
 
 	/*
-	 * Free the projection info and the scan attribute info
-	 *
-	 * Note: we don't ExecFreeResultType(scanstate) because the rule manager
-	 * depends on the tupType returned by ExecMain().  So for now, this is
-	 * freed at end-transaction time.  -cim 6/2/91
+	 * Free the exprcontext(s)
 	 */
-	ExecFreeProjectionInfo(&scanstate->cstate);
-	ExecFreeExprContext(&scanstate->cstate);
-	if (indexstate->iss_RuntimeContext)
-		FreeExprContext(indexstate->iss_RuntimeContext);
+	ExecFreeExprContext(&node->ss.ps);
+	if (node->iss_RuntimeContext)
+		FreeExprContext(node->iss_RuntimeContext);
+
+	/*
+	 * clear out tuple table slots
+	 */
+	ExecClearTuple(node->ss.ps.ps_ResultTupleSlot);
+	ExecClearTuple(node->ss.ss_ScanTupleSlot);
+
+	/* drop hash table */
+	if (node->iss_DupHash)
+		hash_destroy(node->iss_DupHash);
 
 	/*
 	 * close the index relations
@@ -489,37 +541,6 @@ ExecEndIndexScan(IndexScan *node)
 	 * locking, however.)
 	 */
 	heap_close(relation, NoLock);
-
-	/*
-	 * free the scan keys used in scanning the indices
-	 */
-	for (i = 0; i < numIndices; i++)
-	{
-		if (scanKeys[i] != NULL)
-			pfree(scanKeys[i]);
-	}
-	pfree(scanKeys);
-	pfree(numScanKeys);
-
-	if (runtimeKeyInfo)
-	{
-		for (i = 0; i < numIndices; i++)
-		{
-			if (runtimeKeyInfo[i] != NULL)
-				pfree(runtimeKeyInfo[i]);
-		}
-		pfree(runtimeKeyInfo);
-	}
-
-	/*
-	 * clear out tuple table slots
-	 */
-	ExecClearTuple(scanstate->cstate.cs_ResultTupleSlot);
-	ExecClearTuple(scanstate->css_ScanTupleSlot);
-	pfree(scanstate);
-	pfree(indexstate->iss_RelationDescs);
-	pfree(indexstate->iss_ScanDescs);
-	pfree(indexstate);
 }
 
 /* ----------------------------------------------------------------
@@ -531,22 +552,20 @@ ExecEndIndexScan(IndexScan *node)
  * ----------------------------------------------------------------
  */
 void
-ExecIndexMarkPos(IndexScan *node)
+ExecIndexMarkPos(IndexScanState *node)
 {
-	IndexScanState *indexstate;
 	IndexScanDescPtr indexScanDescs;
 	IndexScanDesc scanDesc;
 	int			indexPtr;
 
-	indexstate = node->indxstate;
-	indexPtr = indexstate->iss_MarkIndexPtr = indexstate->iss_IndexPtr;
-	indexScanDescs = indexstate->iss_ScanDescs;
-	scanDesc = indexScanDescs[indexPtr];
+	indexPtr = node->iss_MarkIndexPtr = node->iss_IndexPtr;
+	if (indexPtr >= 0 && indexPtr < node->iss_NumIndices)
+	{
+		indexScanDescs = node->iss_ScanDescs;
+		scanDesc = indexScanDescs[indexPtr];
 
-#ifdef NOT_USED
-	IndexScanMarkPosition(scanDesc);
-#endif
-	index_markpos(scanDesc);
+		index_markpos(scanDesc);
+	}
 }
 
 /* ----------------------------------------------------------------
@@ -555,32 +574,28 @@ ExecIndexMarkPos(IndexScan *node)
  * old comments
  *		Restores scan position by restoring the current index.
  *		Returns nothing.
- *
- *		XXX Assumes previously marked scan position belongs to current index
  * ----------------------------------------------------------------
  */
 void
-ExecIndexRestrPos(IndexScan *node)
+ExecIndexRestrPos(IndexScanState *node)
 {
-	IndexScanState *indexstate;
 	IndexScanDescPtr indexScanDescs;
 	IndexScanDesc scanDesc;
 	int			indexPtr;
 
-	indexstate = node->indxstate;
-	indexPtr = indexstate->iss_IndexPtr = indexstate->iss_MarkIndexPtr;
-	indexScanDescs = indexstate->iss_ScanDescs;
-	scanDesc = indexScanDescs[indexPtr];
+	indexPtr = node->iss_IndexPtr = node->iss_MarkIndexPtr;
+	if (indexPtr >= 0 && indexPtr < node->iss_NumIndices)
+	{
+		indexScanDescs = node->iss_ScanDescs;
+		scanDesc = indexScanDescs[indexPtr];
 
-#ifdef NOT_USED
-	IndexScanRestorePosition(scanDesc);
-#endif
-	index_restrpos(scanDesc);
+		index_restrpos(scanDesc);
+	}
 }
 
 /* ----------------------------------------------------------------
  *		ExecInitIndexScan
-  *
+ *
  *		Initializes the index scan's state information, creates
  *		scan keys, and opens the base and index relations.
  *
@@ -597,11 +612,10 @@ ExecIndexRestrPos(IndexScan *node)
  *		  estate: the execution state initialized in InitPlan.
  * ----------------------------------------------------------------
  */
-bool
-ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
+IndexScanState *
+ExecInitIndexScan(IndexScan *node, EState *estate)
 {
 	IndexScanState *indexstate;
-	CommonScanState *scanstate;
 	List	   *indxqual;
 	List	   *indxid;
 	List	   *listscan;
@@ -612,7 +626,7 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 	int		   *numScanKeys;
 	RelationPtr indexDescs;
 	IndexScanDescPtr scanDescs;
-	int		  **runtimeKeyInfo;
+	ExprState ***runtimeKeyInfo;
 	bool		have_runtime_keys;
 	RangeTblEntry *rtentry;
 	Index		relid;
@@ -620,45 +634,46 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 	Relation	currentRelation;
 
 	/*
-	 * assign execution state to node
+	 * create state structure
 	 */
-	node->scan.plan.state = estate;
-
-	/*
-	 * Part 1)	initialize scan state
-	 *
-	 * create new CommonScanState for node
-	 */
-	scanstate = makeNode(CommonScanState);
-	node->scan.scanstate = scanstate;
+	indexstate = makeNode(IndexScanState);
+	indexstate->ss.ps.plan = (Plan *) node;
+	indexstate->ss.ps.state = estate;
 
 	/*
 	 * Miscellaneous initialization
 	 *
 	 * create expression context for node
 	 */
-	ExecAssignExprContext(estate, &scanstate->cstate);
+	ExecAssignExprContext(estate, &indexstate->ss.ps);
+
+	/*
+	 * initialize child expressions
+	 */
+	indexstate->ss.ps.targetlist = (List *)
+		ExecInitExpr((Expr *) node->scan.plan.targetlist,
+					 (PlanState *) indexstate);
+	indexstate->ss.ps.qual = (List *)
+		ExecInitExpr((Expr *) node->scan.plan.qual,
+					 (PlanState *) indexstate);
+	indexstate->indxqual = (List *)
+		ExecInitExpr((Expr *) node->indxqual,
+					 (PlanState *) indexstate);
+	indexstate->indxqualorig = (List *)
+		ExecInitExpr((Expr *) node->indxqualorig,
+					 (PlanState *) indexstate);
 
 #define INDEXSCAN_NSLOTS 2
 
 	/*
 	 * tuple table initialization
 	 */
-	ExecInitResultTupleSlot(estate, &scanstate->cstate);
-	ExecInitScanTupleSlot(estate, scanstate);
+	ExecInitResultTupleSlot(estate, &indexstate->ss.ps);
+	ExecInitScanTupleSlot(estate, &indexstate->ss);
 
 	/*
-	 * initialize projection info.	result type comes from scan desc
-	 * below..
+	 * Initialize index-specific scan state
 	 */
-	ExecAssignProjectionInfo((Plan *) node, &scanstate->cstate);
-
-	/*
-	 * Part 2)	initialize index scan state
-	 *
-	 * create new IndexScanState for node
-	 */
-	indexstate = makeNode(IndexScanState);
 	indexstate->iss_NumIndices = 0;
 	indexstate->iss_IndexPtr = -1;
 	indexstate->iss_ScanKeys = NULL;
@@ -669,8 +684,6 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 	indexstate->iss_RelationDescs = NULL;
 	indexstate->iss_ScanDescs = NULL;
 
-	node->indxstate = indexstate;
-
 	/*
 	 * get the index node information
 	 */
@@ -694,7 +707,7 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 	 * initialize space for runtime key info (may not be needed)
 	 */
 	have_runtime_keys = false;
-	runtimeKeyInfo = (int **) palloc(numIndices * sizeof(int *));
+	runtimeKeyInfo = (ExprState ***) palloc0(numIndices * sizeof(ExprState **));
 
 	/*
 	 * build the index scan keys from the index qualification
@@ -706,15 +719,15 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 		List	   *qual;
 		int			n_keys;
 		ScanKey		scan_keys;
-		int		   *run_keys;
+		ExprState **run_keys;
 
 		qual = lfirst(indxqual);
 		indxqual = lnext(indxqual);
 		n_keys = length(qual);
 		scan_keys = (n_keys <= 0) ? (ScanKey) NULL :
 			(ScanKey) palloc(n_keys * sizeof(ScanKeyData));
-		run_keys = (n_keys <= 0) ? (int *) NULL :
-			(int *) palloc(n_keys * sizeof(int));
+		run_keys = (n_keys <= 0) ? (ExprState **) NULL :
+			(ExprState **) palloc(n_keys * sizeof(ExprState *));
 
 		CXT1_printf("ExecInitIndexScan: context is %d\n", CurrentMemoryContext);
 
@@ -725,28 +738,26 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 		listscan = qual;
 		for (j = 0; j < n_keys; j++)
 		{
-			Expr	   *clause; /* one clause of index qual */
-			Oper	   *op;		/* operator used in clause */
-			Node	   *leftop; /* expr on lhs of operator */
-			Node	   *rightop;	/* expr on rhs ... */
+			OpExpr	   *clause; /* one clause of index qual */
+			Expr	   *leftop; /* expr on lhs of operator */
+			Expr	   *rightop;	/* expr on rhs ... */
 			bits16		flags = 0;
 
 			int			scanvar;	/* which var identifies varattno */
 			AttrNumber	varattno = 0;	/* att number used in scan */
-			Oid			opid;	/* operator id used in scan */
+			Oid			opfuncid;		/* operator id used in scan */
 			Datum		scanvalue = 0;	/* value used in scan (if const) */
 
 			/*
 			 * extract clause information from the qualification
 			 */
-			clause = lfirst(listscan);
+			clause = (OpExpr *) lfirst(listscan);
 			listscan = lnext(listscan);
 
-			op = (Oper *) clause->oper;
-			if (!IsA(clause, Expr) ||!IsA(op, Oper))
-				elog(ERROR, "ExecInitIndexScan: indxqual not an opclause!");
+			if (!IsA(clause, OpExpr))
+				elog(ERROR, "indxqual is not an OpExpr");
 
-			opid = op->opid;
+			opfuncid = clause->opfuncid;
 
 			/*
 			 * Here we figure out the contents of the index qual. The
@@ -764,9 +775,9 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 			 * which case we need to recalculate the index scan key at run
 			 * time.
 			 *
-			 * Hence, we set have_runtime_keys to true and then set the
-			 * appropriate flag in run_keys to LEFT_OP or RIGHT_OP. The
-			 * corresponding scan keys are recomputed at run time.
+			 * Hence, we set have_runtime_keys to true and place the
+			 * appropriate subexpression in run_keys. The corresponding
+			 * scan key values are recomputed at run time.
 			 *
 			 * XXX Although this code *thinks* it can handle an indexqual
 			 * with the indexkey on either side, in fact it cannot.
@@ -784,19 +795,20 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 			 */
 
 			scanvar = NO_OP;
-			run_keys[j] = NO_OP;
+			run_keys[j] = NULL;
 
 			/*
 			 * determine information in leftop
 			 */
-			leftop = (Node *) get_leftop(clause);
+			leftop = (Expr *) get_leftop((Expr *) clause);
 
 			if (leftop && IsA(leftop, RelabelType))
 				leftop = ((RelabelType *) leftop)->arg;
 
 			Assert(leftop != NULL);
 
-			if (IsA(leftop, Var) &&var_is_rel((Var *) leftop))
+			if (IsA(leftop, Var) &&
+				var_is_rel((Var *) leftop))
 			{
 				/*
 				 * if the leftop is a "rel-var", then it means that it is
@@ -816,32 +828,6 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 				if (((Const *) leftop)->constisnull)
 					flags |= SK_ISNULL;
 			}
-			else if (IsA(leftop, Param))
-			{
-				bool		isnull;
-
-				/*
-				 * if the leftop is a Param node then it means it
-				 * identifies the value to place in our scan key.
-				 */
-
-				/* Life was so easy before ... subselects */
-				if (((Param *) leftop)->paramkind == PARAM_EXEC)
-				{
-					/* treat Param as runtime key */
-					have_runtime_keys = true;
-					run_keys[j] = LEFT_OP;
-				}
-				else
-				{
-					/* treat Param like a constant */
-					scanvalue = ExecEvalParam((Param *) leftop,
-										scanstate->cstate.cs_ExprContext,
-											  &isnull);
-					if (isnull)
-						flags |= SK_ISNULL;
-				}
-			}
 			else
 			{
 				/*
@@ -850,28 +836,28 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 				 * key.
 				 */
 				have_runtime_keys = true;
-				run_keys[j] = LEFT_OP;
+				run_keys[j] = ExecInitExpr(leftop, (PlanState *) indexstate);
 			}
 
 			/*
 			 * now determine information in rightop
 			 */
-			rightop = (Node *) get_rightop(clause);
+			rightop = (Expr *) get_rightop((Expr *) clause);
 
 			if (rightop && IsA(rightop, RelabelType))
 				rightop = ((RelabelType *) rightop)->arg;
 
 			Assert(rightop != NULL);
 
-			if (IsA(rightop, Var) &&var_is_rel((Var *) rightop))
+			if (IsA(rightop, Var) &&
+				var_is_rel((Var *) rightop))
 			{
 				/*
 				 * here we make sure only one op identifies the
 				 * scan-attribute...
 				 */
 				if (scanvar == LEFT_OP)
-					elog(ERROR, "ExecInitIndexScan: %s",
-						 "both left and right op's are rel-vars");
+					elog(ERROR, "both left and right operands are rel-vars");
 
 				/*
 				 * if the rightop is a "rel-var", then it means that it is
@@ -891,32 +877,6 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 				if (((Const *) rightop)->constisnull)
 					flags |= SK_ISNULL;
 			}
-			else if (IsA(rightop, Param))
-			{
-				bool		isnull;
-
-				/*
-				 * if the rightop is a Param node then it means it
-				 * identifies the value to place in our scan key.
-				 */
-
-				/* Life was so easy before ... subselects */
-				if (((Param *) rightop)->paramkind == PARAM_EXEC)
-				{
-					/* treat Param as runtime key */
-					have_runtime_keys = true;
-					run_keys[j] = RIGHT_OP;
-				}
-				else
-				{
-					/* treat Param like a constant */
-					scanvalue = ExecEvalParam((Param *) rightop,
-										scanstate->cstate.cs_ExprContext,
-											  &isnull);
-					if (isnull)
-						flags |= SK_ISNULL;
-				}
-			}
 			else
 			{
 				/*
@@ -925,7 +885,7 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 				 * key.
 				 */
 				have_runtime_keys = true;
-				run_keys[j] = RIGHT_OP;
+				run_keys[j] = ExecInitExpr(rightop, (PlanState *) indexstate);
 			}
 
 			/*
@@ -933,8 +893,7 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 			 * attribute...
 			 */
 			if (scanvar == NO_OP)
-				elog(ERROR, "ExecInitIndexScan: %s",
-					 "neither leftop nor rightop refer to scan relation");
+				elog(ERROR, "neither left nor right operand refer to scan relation");
 
 			/*
 			 * initialize the scan key's fields appropriately
@@ -943,7 +902,7 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 								   flags,
 								   varattno,	/* attribute number to
 												 * scan */
-								   (RegProcedure) opid, /* reg proc to use */
+								   opfuncid,	/* reg proc to use */
 								   scanvalue);	/* constant */
 		}
 
@@ -976,12 +935,12 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 	 */
 	if (have_runtime_keys)
 	{
-		ExprContext *stdecontext = scanstate->cstate.cs_ExprContext;
+		ExprContext *stdecontext = indexstate->ss.ps.ps_ExprContext;
 
-		ExecAssignExprContext(estate, &scanstate->cstate);
+		ExecAssignExprContext(estate, &indexstate->ss.ps);
 		indexstate->iss_RuntimeKeyInfo = runtimeKeyInfo;
-		indexstate->iss_RuntimeContext = scanstate->cstate.cs_ExprContext;
-		scanstate->cstate.cs_ExprContext = stdecontext;
+		indexstate->iss_RuntimeContext = indexstate->ss.ps.ps_ExprContext;
+		indexstate->ss.ps.ps_ExprContext = stdecontext;
 	}
 	else
 	{
@@ -1005,17 +964,13 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 
 	currentRelation = heap_open(reloid, AccessShareLock);
 
-	if (!RelationGetForm(currentRelation)->relhasindex)
-		elog(ERROR, "indexes of the relation %u was inactivated", reloid);
-
-	scanstate->css_currentRelation = currentRelation;
-	scanstate->css_currentScanDesc = NULL;		/* no heap scan here */
+	indexstate->ss.ss_currentRelation = currentRelation;
+	indexstate->ss.ss_currentScanDesc = NULL;	/* no heap scan here */
 
 	/*
 	 * get the scan type from the relation descriptor.
 	 */
-	ExecAssignScanType(scanstate, RelationGetDescr(currentRelation), false);
-	ExecAssignResultTypeFromTL((Plan *) node, &scanstate->cstate);
+	ExecAssignScanType(&indexstate->ss, RelationGetDescr(currentRelation), false);
 
 	/*
 	 * open the index relations and initialize relation and scan
@@ -1026,7 +981,7 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 	listscan = indxid;
 	for (i = 0; i < numIndices; i++)
 	{
-		Oid			indexOid = (Oid) lfirsti(listscan);
+		Oid			indexOid = lfirsto(listscan);
 
 		indexDescs[i] = index_open(indexOid);
 		scanDescs[i] = index_beginscan(currentRelation,
@@ -1040,10 +995,46 @@ ExecInitIndexScan(IndexScan *node, EState *estate, Plan *parent)
 	indexstate->iss_RelationDescs = indexDescs;
 	indexstate->iss_ScanDescs = scanDescs;
 
+	/*
+	 * Initialize result tuple type and projection info.
+	 */
+	ExecAssignResultTypeFromTL(&indexstate->ss.ps);
+	ExecAssignScanProjectionInfo(&indexstate->ss);
+
+	/*
+	 * Initialize hash table if needed.
+	 */
+	if (numIndices > 1)
+		create_duphash(indexstate);
+	else
+		indexstate->iss_DupHash = NULL;
+
 	/*
 	 * all done.
 	 */
-	return TRUE;
+	return indexstate;
+}
+
+static void
+create_duphash(IndexScanState *node)
+{
+	HASHCTL		hash_ctl;
+
+	MemSet(&hash_ctl, 0, sizeof(hash_ctl));
+	hash_ctl.keysize = SizeOfIptrData;
+	hash_ctl.entrysize = sizeof(DupHashTabEntry);
+	hash_ctl.hash = tag_hash;
+	hash_ctl.hcxt = CurrentMemoryContext;
+	node->iss_DupHash = hash_create("DupHashTable",
+									(long) ceil(node->ss.ps.plan->plan_rows),
+									&hash_ctl,
+									HASH_ELEM | HASH_FUNCTION | HASH_CONTEXT);
+	if (node->iss_DupHash == NULL)
+		ereport(ERROR,
+				(errcode(ERRCODE_OUT_OF_MEMORY),
+				 errmsg("out of memory")));
+	node->iss_MaxHash = (SortMem * 1024L) /
+		(MAXALIGN(sizeof(HASHELEMENT)) + MAXALIGN(sizeof(DupHashTabEntry)));
 }
 
 int
