@@ -3,14 +3,14 @@
  * fe-auth.c
  *	   The front-end (client) authorization routines
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * NOTE: the error message strings returned by this module must not
  * exceed INITIAL_EXPBUFFER_SIZE (currently 256 bytes).
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/interfaces/libpq/fe-auth.c,v 1.71 2002/09/04 20:31:46 momjian Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/interfaces/libpq/fe-auth.c,v 1.84 2003/10/25 03:48:47 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -270,7 +270,7 @@ pg_an_to_ln(char *aname)
 
 
 /*
- * Various krb5 state which is not connection specfic, and a flag to
+ * Various krb5 state which is not connection specific, and a flag to
  * indicate whether we have initialised it yet.
  */
 static int	pg_krb5_initialised;
@@ -357,10 +357,7 @@ pg_krb5_authname(char *PQerrormsg)
  *					   the server
  */
 static int
-pg_krb5_sendauth(char *PQerrormsg, int sock,
-				 struct sockaddr_in * laddr,
-				 struct sockaddr_in * raddr,
-				 const char *hostname)
+pg_krb5_sendauth(char *PQerrormsg, int sock, const char *hostname)
 {
 	krb5_error_code retval;
 	int			ret;
@@ -391,8 +388,10 @@ pg_krb5_sendauth(char *PQerrormsg, int sock,
 	flags = fcntl(sock, F_GETFL);
 	if (flags < 0 || fcntl(sock, F_SETFL, (long) (flags & ~O_NONBLOCK)))
 	{
+		char		sebuf[256];
+
 		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
-				 libpq_gettext("could not set socket to blocking mode: %s\n"), strerror(errno));
+				 libpq_gettext("could not set socket to blocking mode: %s\n"), pqStrerror(errno, sebuf, sizeof(sebuf)));
 		krb5_free_principal(pg_krb5_context, server);
 		return STATUS_ERROR;
 	}
@@ -436,9 +435,11 @@ pg_krb5_sendauth(char *PQerrormsg, int sock,
 
 	if (fcntl(sock, F_SETFL, (long) flags))
 	{
+		char		sebuf[256];
+
 		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 				 libpq_gettext("could not restore non-blocking mode on socket: %s\n"),
-				 strerror(errno));
+				 pqStrerror(errno, sebuf, sizeof(sebuf)));
 		ret = STATUS_ERROR;
 	}
 
@@ -449,7 +450,9 @@ pg_krb5_sendauth(char *PQerrormsg, int sock,
 static int
 pg_local_sendauth(char *PQerrormsg, PGconn *conn)
 {
-#if defined(HAVE_STRUCT_CMSGCRED) || defined(HAVE_STRUCT_FCRED) || (defined(HAVE_STRUCT_SOCKCRED) && defined(LOCAL_CREDS))
+#if defined(HAVE_STRUCT_CMSGCRED) || defined(HAVE_STRUCT_FCRED) || \
+	(defined(HAVE_STRUCT_SOCKCRED) && defined(LOCAL_CREDS)) && \
+	!defined(HAVE_GETPEEREID) && !defined(SO_PEERCRED)
 	char		buf;
 	struct iovec iov;
 	struct msghdr msg;
@@ -461,13 +464,6 @@ pg_local_sendauth(char *PQerrormsg, PGconn *conn)
 	/* Point to start of first structure */
 	struct cmsghdr *cmsg = (struct cmsghdr *) cmsgmem;
 #endif
-#ifdef HAVE_STRUCT_SOCKCRED
-	/* Prevent padding */
-	char		cmsgmem[sizeof(struct cmsghdr) + sizeof(struct sockcred)];
-
-	/* Point to start of first structure */
-	struct cmsghdr *cmsg = (struct cmsghdr *) cmsgmem;
-#endif
 
 	/*
 	 * The backend doesn't care what we send here, but it wants exactly
@@ -493,8 +489,11 @@ pg_local_sendauth(char *PQerrormsg, PGconn *conn)
 
 	if (sendmsg(conn->sock, &msg, 0) == -1)
 	{
+		char		sebuf[256];
+
 		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
-				 "pg_local_sendauth: sendmsg: %s\n", strerror(errno));
+				 "pg_local_sendauth: sendmsg: %s\n",
+				 pqStrerror(errno, sebuf, sizeof(sebuf)));
 		return STATUS_ERROR;
 	}
 	return STATUS_OK;
@@ -557,7 +556,11 @@ pg_password_sendauth(PGconn *conn, const char *password, AuthRequest areq)
 		default:
 			return STATUS_ERROR;
 	}
-	ret = pqPacketSend(conn, crypt_pwd, strlen(crypt_pwd) + 1);
+	/* Packet has a message type as of protocol 3.0 */
+	if (PG_PROTOCOL_MAJOR(conn->pversion) >= 3)
+		ret = pqPacketSend(conn, 'p', crypt_pwd, strlen(crypt_pwd) + 1);
+	else
+		ret = pqPacketSend(conn, 0, crypt_pwd, strlen(crypt_pwd) + 1);
 	if (areq == AUTH_REQ_MD5)
 		free(crypt_pwd);
 	return ret;
@@ -581,8 +584,9 @@ fe_sendauth(AuthRequest areq, PGconn *conn, const char *hostname,
 
 		case AUTH_REQ_KRB4:
 #ifdef KRB4
-			if (pg_krb4_sendauth(PQerrormsg, conn->sock, &conn->laddr.in,
-								 &conn->raddr.in,
+			if (pg_krb4_sendauth(PQerrormsg, conn->sock,
+							   (struct sockaddr_in *) & conn->laddr.addr,
+							   (struct sockaddr_in *) & conn->raddr.addr,
 								 hostname) != STATUS_OK)
 			{
 				snprintf(PQerrormsg, PQERRORMSG_LENGTH,
@@ -598,8 +602,7 @@ fe_sendauth(AuthRequest areq, PGconn *conn, const char *hostname,
 
 		case AUTH_REQ_KRB5:
 #ifdef KRB5
-			if (pg_krb5_sendauth(PQerrormsg, conn->sock, &conn->laddr.in,
-								 &conn->raddr.in,
+			if (pg_krb5_sendauth(PQerrormsg, conn->sock,
 								 hostname) != STATUS_OK)
 			{
 				snprintf(PQerrormsg, PQERRORMSG_LENGTH,
@@ -684,7 +687,14 @@ MsgType
 fe_getauthsvc(char *PQerrormsg)
 {
 	if (pg_authsvc < 0 || pg_authsvc >= n_authsvcs)
+	{
 		fe_setauthsvc(DEFAULT_CLIENT_AUTHSVC, PQerrormsg);
+		if (pg_authsvc < 0 || pg_authsvc >= n_authsvcs)
+		{
+			/* Can only get here if DEFAULT_CLIENT_AUTHSVC is misdefined */
+			return 0;
+		}
+	}
 	return authsvcs[pg_authsvc].msgtype;
 }
 
@@ -702,6 +712,10 @@ fe_getauthname(char *PQerrormsg)
 
 	authsvc = fe_getauthsvc(PQerrormsg);
 
+	/* this just guards against broken DEFAULT_CLIENT_AUTHSVC, see above */
+	if (authsvc == 0)
+		return NULL;			/* leave original error message in place */
+
 #ifdef KRB4
 	if (authsvc == STARTUP_KRB4_MSG)
 		name = pg_krb4_authname(PQerrormsg);
@@ -722,9 +736,12 @@ fe_getauthname(char *PQerrormsg)
 		if (GetUserName(username, &namesize))
 			name = username;
 #else
-		struct passwd *pw = getpwuid(geteuid());
+		char		pwdbuf[BUFSIZ];
+		struct passwd pwdstr;
+		struct passwd *pw = NULL;
 
-		if (pw)
+		if (pqGetpwuid(geteuid(), &pwdstr,
+					   pwdbuf, sizeof(pwdbuf), &pw) == 0)
 			name = pw->pw_name;
 #endif
 	}
